# CanopyView 2D Plugins - schematic.py
# Description: Schematic plugin.
# Contributors: Nik Molnar
# Created: 2009-06-17
# (c)2009 The Evergreen State College

import math
from PIL import Image
from Plugins import *
from OmniImage import *

#Hard-coded options
kPixelsToMeter		= 10 #Number of pixels in one meter
kCellSize			= 50 #Size of one grid square in pixels
kBaseImageSize		= (300,875) #Should be changed later to take user-input or use the largest tree in the dataset.

def init():
	return SchematicPlugin()
	
class SchematicPlugin:
	"Schematic plugin class."
	
	def __init__(self):
		"Constructor."
		
		self.dataset		= None
		self.stemsById		= { }
		
		
	def GetStem(self, dataset, siteId, stemId):
		"Internal method. Get a single tree and associated branches for the given dataset."
		
		#Resolve column and table names
		nameStem			= dataset.MapTable("stem")
		nameBranch			= dataset.MapTable("branch")
		nameReiteration		= dataset.MapTable("reiteration")
		nameSiteid			= dataset.MapColumn("site_id")
		nameStemid			= dataset.MapColumn("stem_id")
		nameDbh				= dataset.MapColumn("dbh")
		nameLivestatus		= dataset.MapColumn("live_status")
		nameHeight			= dataset.MapColumn("height")
		nameBranchdata		= dataset.MapColumn("branch_data?")
		nameBranchid		= dataset.MapColumn("branch_id")
		nameDiameter		= dataset.MapColumn("diameter")
		nameAzimuth			= dataset.MapColumn("azimuth")
		nameExtent			= dataset.MapColumn("extent")
		nameSlope			= dataset.MapColumn("slope")
		nameEpicormic		= dataset.MapColumn("epicormic")
		nameReiterationid	= dataset.MapColumn("reiteration_id")
		nameBaseht			= dataset.MapColumn("base_ht")
		nameTopht			= dataset.MapColumn("top_ht")
		nameBaseazi			= dataset.MapColumn("base azi")
		nameTopazi			= dataset.MapColumn("top_azi")
		nameBasaldiameter	= dataset.MapColumn("basal_diameter")
		nameBasedist		= dataset.MapColumn("base dist")
		nameTopdist			= dataset.MapColumn("top_dist")
		
		result	= dataset.ExecQuery("SELECT * FROM %s WHERE %s = \"%s\" AND %s = %d" % (
					nameStem, nameSiteid, siteId, nameStemid, stemId))
		row		= result.NextRowD()
		
		if not row:
			return None
			
		stem = Stem()
		if row[nameStemid]:		stem.id 			= int(row[nameStemid])
		if row[nameSiteid]:		stem.site			= row[nameSiteid]
		if row[nameHeight]:		stem.height			= float(row[nameHeight])
		if row[nameDbh]:		stem.diameter		= float(row[nameDbh])
		if row[nameLivestatus]:	stem.alive			= (row[nameLivestatus].lower() == "l")
		if row[nameBranchdata]:	stem.hasBranchData	= (row[nameBranchdata] == "1" or 
			row[nameBranchdata] == True)
		
		result = dataset.ExecQuery("SELECT * FROM %s WHERE %s = \"%s\" AND %s = %d" % 
				(nameBranch, nameSiteid, siteId, nameStemid, stemId))
				
		while 1:
			row = result.NextRowD()
			
			if not row:
				break
				
			branch = Branch()
			if row[nameBranchid]:		branch.id			= int(row[nameBranchid])
			if row[nameStemid]:			branch.stem			= int(row[nameStemid])
			if row[nameSiteid]:			branch.site			= row[nameSiteid]
			if row[nameHeight]:			branch.height		= float(row[nameHeight])
			if row[nameExtent]:			branch.extent		= float(row[nameExtent])
			if row[nameAzimuth]:		branch.azimuth		= int(row[nameAzimuth])
			if row[nameDiameter]:		branch.diameter		= float(row[nameDiameter]) / 100
			if row[nameLivestatus]:		branch.alive		= (row[nameLivestatus].lower() == "l")
			if row[nameSlope]:			branch.slope		= int(row[nameSlope])
			if row[nameEpicormic]:		branch.epicormic	= (row[nameEpicormic] == "epicormic")
			
			stem.branches.append(branch)
			
		result = dataset.ExecQuery( "SELECT * FROM %s WHERE %s = \"%s\" AND %s = %d" % 
					(nameReiteration, nameSiteid, siteId, nameStemid, stemId))

		while 1:
			row = result.NextRowD()
			
			if not row:
				break
				
			reit = Reiteration()
			if row[nameReiterationid]:	reit.id			= int(row[nameReiterationid])
			if row[nameStemid]:			reit.stem		= int(row[nameStemid])
			if row[nameSiteid]:			reit.site		= row[nameSiteid]
			if row[nameBaseazi]:		reit.azimuth	= int(row[nameBaseazi])
			if row[nameBaseht]:			reit.base		= float(row[nameBaseht])
			if row[nameTopht]:			reit.top		= float(row[nameTopht])
			if row[nameBasaldiameter]:	reit.diameter	= float(row[nameBasaldiameter]) / 100
			if row[nameBasedist]:		reit.distance	= float(row[nameBasedist])
			
			if row[nameTopazi] and not row[nameBaseazi]:	reit.azimuth	= int(row[nameTopazi])
			if row[nameTopdist] and not row[nameBasedist]:	reit.distance	= float(row[nameTopdist])
			
			stem.reiterations.append(reit)
		
		return stem
		
	def GetInverseColor(self, color):
		"Internal method. Get an inverse of an RGB color"
		
		return (abs(color[0]-255), abs(color[1]-255), abs(color[2]-255))
	
	def Info(self):
		"Plugin information."
		
		info = {
			'name'			: "Branch Schematic",
			'description'	: "Generates 2D images representing branch deatils, including epicormics and reiterations.",
			'unit_label'	: "Stem",
			'views'			: 1
		}
		
		return info
		
	def RequiredType(self):
		"Returns 'sql', 'flat', or None depending on required dataset type."
		
		return 'sql'
		
	def RequiredTables(self):
		"Returns a list of required tables, or None if expecting a table-less dataset."
		
		return ["stem", "branch", "reiteration"]
		
	def RequiredColumns(self):
		"Returns a list of required columns in the form [(table, column), ...]"

		return [("reiteration", "reiteration_id"), ("reiteration", "stem_id"), 
				("reiteration", "site_id"), ("reiteration", "base_ht"), ("reiteration", "top_ht"),
				("reiteration", "base azi"), ("reiteration", "top_azi"), 
				("reiteration", "basal_diameter"), ("reiteration", "base dist"), 
				("reiteration", "top_dist"), ("stem", "site_id"), ("stem", "stem_id"),
				("stem", "dbh"), ("stem", "height"), ("stem", "branch_data?"), 
				("stem", "live_status"), ("branch", "site_id"), ("branch", "stem_id"),
				("branch", "branch_id"), ("branch", "live_status"), ("branch", "height"),
				("branch", "diameter"), ("branch", "azimuth"), ("branch", "extent"),
				("branch", "slope"), ("branch", "epicormic")]
		
	def ValidData(self, dataset):
		'''Returns True if the provided dataset contains the data necessary to generate this visualization. 
		   False otherwise.'''
		
		#Check Tables
		if not dataset.HasTable("stem", True):
			return False
			
		if not dataset.HasTable("branch", True):
			return False
			
		if not dataset.HasTable("reiteration", True):
			return False
		
		#Reiteration Table
		if not dataset.HasColumn("reiteration", "reiteration_id", True):
			return False
			
		if not dataset.HasColumn("reiteration", "stem_id", True):
			return False
			
		if not dataset.HasColumn("reiteration", "site_id", True):
			return False
			
		if not dataset.HasColumn("reiteration", "base_ht", True):
			return False
			
		if not dataset.HasColumn("reiteration", "top_ht", True):
			return False
			
		if not dataset.HasColumn("reiteration", "base azi", True):
			return False
			
		if not dataset.HasColumn("reiteration", "top_azi", True):
			return False
			
		if not dataset.HasColumn("reiteration", "basal_diameter", True):
			return False
			
		if not dataset.HasColumn("reiteration", "base dist", True):
			return False
			
		if not dataset.HasColumn("reiteration", "top_dist", True):
			return False
		
		#Stem Table
		if not dataset.HasColumn("stem", "site_id", True):
			return False
			
		if not dataset.HasColumn("stem", "stem_id", True):
			return False
			
		if not dataset.HasColumn("stem", "dbh", True):
			return False
			
		if not dataset.HasColumn("stem", "height", True):
			return False
			
		if not dataset.HasColumn("stem", "branch_data?", True):
			return False
			
		if not dataset.HasColumn("stem", "live_status", True):
			return False
			
		#Branch Table
		if not dataset.HasColumn("branch", "site_id", True):
			return False
			
		if not dataset.HasColumn("branch", "branch_id", True):
			return False
			
		if not dataset.HasColumn("branch", "stem_id", True):
			return False
			
		if not dataset.HasColumn("branch", "live_status", True):
			return False
			
		if not dataset.HasColumn("branch", "height", True):
			return False
			
		if not dataset.HasColumn("branch", "diameter", True):
			return False
			
		if not dataset.HasColumn("branch", "azimuth", True):
			return False
			
		if not dataset.HasColumn("branch", "extent", True):
			return False
			
		if not dataset.HasColumn("branch", "slope", True):
			return False
			
		if not dataset.HasColumn("branch", "epicormic", True):
			return False
			
		#All There?
		return True
		
	def GetUnits(self, dataset):
		"Returns a list of stems that can be visualized given the available data."
		
		table	= dataset.GetTable("stem")
		units	= [ ]
		
		while 1:
			stem = table.NextRowD()
			
			if stem == None:
				break
				
			if stem['live_status'] == "L" and stem['branch_data?'] == 1:
				try:
					units.append([(stem['site_id'], int(stem['stem_id'])), str(stem['site_id']) + " " + 
						str(stem['stem_id'])])
				except ValueError:
					pass
					
		return units
				
	def GetOptions(self, dataset):
		"Returns a list of options to be provided to the end user at run time."
		
		options = [
			CheckboxOption('grid', "Show Grid", True),
			CheckboxOption('rulers', "Show Rulers", False),
			CheckboxOption('info', "Show Tree Information", True),
			CheckboxOption('hide_center', "Hide Center Line", False),
			CheckboxOption('regular', "Show Regular Branches", True),
			CheckboxOption('epicormic', "Show Epicormic Branches", True),
			CheckboxOption('reiteration', "Show Reiterations", True),
			CheckboxOption('dead', "Show Dead Branches", True),
			SpacerOption(),
			ColorOption('bg_color', "Background Color", (255,255,255)),
			ColorOption('label_color', "Label Background Color", (0,0,0)),
			SpacerOption(),
			ColorOption('live_large_color', "Large Live Branch Color", (175,255,175)),
			ColorOption('live_small_color', "Small Live Branch Color", (0, 255, 0)),
			ColorOption('dead_large_color', "Large Dead Branch Color", (255,127,0)),
			ColorOption('dead_small_color', "Small Dead Branch Color", (155,102,51)),
			ColorOption('epicormic_color', "Epicormic Color", (127, 0, 127)),
			ColorOption('reiteration_color', "Reiteration Color", (0,0,255))
		]
				
		return options
	
	def GetUnitName(self, dataset, options, id, view=0):
		"Returns a filename for the given unit id and view with current dataset and options."
		
		return "Tree %d (%s)" % (id[1], id[0])
	
	def GetImage(self, dataset, options, id, view=0):
		"Returns an OmniImage instance representing the requested stem and view number."
		
		stem = self.GetStem(dataset, id[0], id[1])
		
		if not stem:
			return OmniImage(10,10)
			
		offset = [0,0]
		
		if options['rulers']:
			offset[0] += 50
			
		if options['info']:
			offset[1] += 30
			
		image = OmniImage((kBaseImageSize[0]+offset[0], kBaseImageSize[1]+offset[1]), options['bg_color'])

		#Draw the grid
		if options['grid']:
			#Horizontal Lines
			for i in range(0, int(kBaseImageSize[1]/kCellSize)+1):
				y = kBaseImageSize[1] - (i*kCellSize)
				image.Draw(OI_Line((0 + offset[0], y), (kBaseImageSize[0] + offset[0], y), 1, 
					self.GetInverseColor(options['bg_color'])))
				
			#Vertical Lines
			vLinesRange = range(0, int(kBaseImageSize[0]/kCellSize)+1)
			for i in vLinesRange:
				if not options['hide_center'] or len(vLinesRange)%2 == 0 or i != int(len(vLinesRange)/2):
					x = offset[0] + (i*kCellSize)
					image.Draw(OI_Line((x, 0), (x, kBaseImageSize[1]), 1, self.GetInverseColor(options['bg_color'])))
					
		#Draw the ruler
		if options['rulers']:
			for i in range(1, int(kBaseImageSize[1]/kCellSize)+1):
				y			= kBaseImageSize[1] - (i*kCellSize)
				label		= str(i * kCellSize/kPixelsToMeter) + "m"
				labelSize	= OI_String((0,0), label, 18).GetStringSize()
				corner		= (47 - labelSize[0], y - labelSize[1])
				
				image.Draw(OI_String(corner, label, 18, self.GetInverseColor(options['bg_color'])))
				
		#Draw tree information
		if options['info']:
			#Box
			start	= (offset[0], kBaseImageSize[1] - 30 + offset[1])
			end		= (kBaseImageSize[0] + offset[0], kBaseImageSize[1] + offset[1])
			
			image.Draw(OI_Rectangle(start, end, options['label_color'], line=1))
			
			#Label
			label		= "Tree %s (%s)" % (stem.id, stem.site)
			labelSize	= OI_String((0,0), label, 18).GetStringSize()
			corner		= (int(kBaseImageSize[0]/2 - labelSize[0]/2 + offset[0]), int(kBaseImageSize[1] - 15 - 
							labelSize[1]/2 + offset[1]))
			
			image.Draw(OI_String(corner, label, 18, self.GetInverseColor(options['label_color'])))
			
		arc			= 180
		rightCenter	= 180
		leftCenter	= 0
		rightSide	= (90,270)
		midPoint	= kBaseImageSize[0]/2 + offset[0]
		
		#Draw the stem
		baseRadius	= stem.diameter / 100 / 2 * kPixelsToMeter
		treeHeight	= stem.height * kPixelsToMeter
		
		trunkPoly = [(midPoint - baseRadius, kBaseImageSize[1]), (midPoint + baseRadius, kBaseImageSize[1]), 
						(midPoint + 1, kBaseImageSize[1] - treeHeight), (midPoint - 1, kBaseImageSize[1] - treeHeight)]
		image.Draw(OI_Polygon(trunkPoly, (182, 61, 54)))
		
		#Draw Branches
		for branch in stem.branches:
			if not options['regular'] and branch.alive and not branch.epicormic:
				continue
				
			if branch.epicormic and not options['epicormic']:
				continue
				
			if not branch.alive and not options['dead']:
				continue
			
			if branch.alive:
				if branch.diameter >= .1:
					color = options['live_large_color']
				else:
					color = options['live_small_color']
			else:
				if branch.diameter >= .1:
					color = options['dead_large_color']
				else:
					color = options['dead_small_color']
					
			if branch.epicormic:
				line		= 1
				lineColor	= options['epicormic_color']
			else:
				line		= 0
				lineColor	= (0,0,0)
			
			slopeGain = math.atan(branch.slope * math.pi / 180.0) * branch.extent
			if branch.azimuth > rightSide[0] and branch.azimuth < rightSide[1]:
				extent = branch.extent * -1
			else:
				extent = branch.extent
						
			points = [(midPoint, kBaseImageSize[1] - branch.height*kPixelsToMeter), (midPoint, 
						kBaseImageSize[1] - (branch.height-branch.diameter)*kPixelsToMeter), 
						(midPoint + extent*kPixelsToMeter, kBaseImageSize[1] - (branch.height + branch.diameter/2 + 
						slopeGain) * kPixelsToMeter)]

			image.Draw(OI_Polygon(points, fill=color, line=line, lineColor=lineColor))
			
		#Draw Reiterations
		if options['reiteration']:
			for reit in stem.reiterations:
				if reit.azimuth > rightSide[0] and reit.azimuth < rightSide[1]:
					distance	= reit.distance * -1
					diameter	= reit.diameter * -1
				else:
					distance	= reit.distance
					diameter	= reit.diameter

				points = [(midPoint + distance*kPixelsToMeter, kBaseImageSize[1] - reit.base*kPixelsToMeter),
							(midPoint + distance*kPixelsToMeter + diameter*kPixelsToMeter, kBaseImageSize[1] -
							reit.base*kPixelsToMeter), (midPoint + distance*kPixelsToMeter + diameter/2*kPixelsToMeter,
							kBaseImageSize[1] - reit.top*kPixelsToMeter)]
							
				image.Draw(OI_Polygon(points, options['reiteration_color']))
				image.Draw(OI_Rectangle((midPoint, kBaseImageSize[1] - reit.base*kPixelsToMeter), (midPoint + 
					distance*kPixelsToMeter + diameter*kPixelsToMeter, kBaseImageSize[1] - 
					(reit.base-reit.diameter)*kPixelsToMeter), options['reiteration_color']))

		return image
		
class Stem:
	"Tree struct."
	
	def __init__(self):
		"Constructor."
		
		self.id				= 0
		self.site			= ""
		self.height			= 0.0
		self.diameter		= 0.0
		self.alive			= False
		self.hasBranchData	= False
		self.branches		= [ ]
		self.reiterations	= [ ]
		
class Branch:
	"Branch struct."
	
	def __init__(self):
		"Constructor."
		
		self.id				= 0
		self.stem			= 0
		self.site			= ""
		self.height			= 0.0
		self.extent			= 0.0
		self.azimuth		= 0
		self.diameter		= 0.0
		self.alive			= False
		self.percentCover	= 0
		self.slope			= 0
		self.epicormic		= False
		
class Reiteration:
	"Reiteration struct."
	
	def __init__(self):
		"Constructor."
		
		self.id			= 0
		self.stem		= 0
		self.site		= ""
		self.azimuth	= 0
		self.base		= 0.0
		self.top		= 0.0
		self.diameter	= 0.0
		self.distance	= 0.0
