from Plugins import *
from OmniImage import *

def init():
	return TestPlugin()

class TestPlugin:
	"Test plugin class."
		
	def Info(self):
		"Plugin information."
		
		info = { }
		info['name']		= "Plugin Test"
		info['description']	= "Simple 'visualization' to test the plugin manager."
		info['unit_label']	= "Box"
		info['views']		= 4
		
		return info
		
	def RequiredType(self):
		"Returns 'sql', 'flat', or None depending on required dataset type."
		
		return None
		
	def RequiredTables(self):
		"Returns a list of required tables, or None if expecting a table-less dataset."
		
		return ["stem"]
		
	def RequiredColumns(self):
		"Returns a list of required columns in the form [(table, column), ...]"
		
		return [("stem", "stem_id")]
		
	def ValidData(self, dataset):
		"Returns True if the provided dataset contains the data necessary to generate this visualization. False otherwise."
		
		return dataset.HasColumn("stem", "stem_id", True)
		
	def GetUnits(self, dataset):
		"Returns a list of ids for units from the dataset that can be visualized given the available data."
		
		return [[1, "1"],[2, "2"],[3, "3"]]
		
	def GetOptions(self, dataset):
		"Returns a list of options to be provided to the end user at run time."
		
		options = [
			TextOption('view_label', "View Label", "View"),
			LabelOption("Test label."),
			TextOption('test_text', "Test Text Option just altered", "Default value."),
			NumberOption('test_num', "Test Number Option", 5),
			SliderOption('slider', "Opacity Slider", 0, 100, 100),
			LabelOption("Before spacer."),
			SpacerOption(),
			LabelOption("After spacer."),
			CheckboxOption('test_checkbox', "Checkbox Test"),
			CheckboxOption('test2_checkbox', "Another checkbox", True),
			RadioButtonsOption('test_radio', "Some Radio Buttons", ["One", "Two", "Three"]),
			ColorOption('test_color', "Color Test")
		]
		
		return options
		
	def GetUnitName(self, dataset, options, id, view=0):
		"Returns a filename for the given unit id and view with current dataset and options."
		
		return "Unit %d View %d" % (id, view+1)
		
	def GetImage(self, dataset, options, id, view=0):
		"Returns an OmniImage instance representing the given view for the unit specified by id given the provided data and options."
		
		opacity = options['slider'] * 255 / 100
		
		Log(self, "Test.")
		
		viewLabel = "View"
		if options.has_key('view_label'):
			viewLabel = options['view_label']
		
		if view == 0:
			lineColor = (0,0,0,opacity)
		
		elif view == 1:
			lineColor = (255,0,0,opacity)
			
		elif view == 2:
			lineColor = (0,255,0,opacity)
			
		else:
			lineColor = (0,0,255,opacity)
		
		image = OmniImage((200,200), (255,255,255))
		
		image.Draw(OI_Line((0,0), (200,200), 1, color=lineColor))
		image.Draw(OI_Line((200,0), (0,200), 1, color=lineColor))
		image.Draw(OI_String((10,10), "%s %d" % (viewLabel, view)))
		
		return image